import json
from functools import wraps

import django_nyt
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse


def disable_notify(f):
    """Disable notifications. Example:

    @disable_notify
    def your_function():
        notify("no one will be notified", ...)
    """
    @wraps(f)
    def wrapper(request, *args, **kwargs):
        django_nyt._disable_notifications = True
        response = f(request, *args, **kwargs)
        django_nyt._disable_notifications = False
        return response
    return wrapper


def login_required_ajax(f):
    """Similar to login_required. But if the request is an ajax request, then
    it returns an error in json with a 403 status code."""

    @wraps(f)
    def wrapper(request, *args, **kwargs):
        if request.is_ajax():
            if not request.user or not request.user.is_authenticated:
                return json_view(lambda *a,
                                 **kw: {'error': 'not logged in'})(request,
                                                                   status=403)
            return f(request, *args, **kwargs)
        else:
            return login_required(f)(request, *args, **kwargs)
    return wrapper


def data2jsonresponse(data, **kwargs):
    json_data = json.dumps(data, ensure_ascii=False)
    status = kwargs.get('status', 200)
    response = HttpResponse(content_type='application/json', status=status)
    response.write(json_data)
    return response


def json_view(f):
    @wraps(f)
    def wrapper(request, *args, **kwargs):
        data = f(request, *args, **kwargs)
        return data2jsonresponse(data, **kwargs)
    return wrapper



from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('django_nyt', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Settings', fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, verbose_name='ID', serialize=False, primary_key=True
                    )
                ), (
                    'user',
                    models.ForeignKey(
                        to_field='id', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
                ), (
                    'interval',
                    models.SmallIntegerField(
                        choices=[
                            (0, 'instantly'), (1380, 'daily'), (9660, 'weekly')
                        ],
                        default=0,
                        verbose_name='interval')
                ),
            ], options={
                'verbose_name': 'settings', 'db_table': 'nyt_settings', 'verbose_name_plural': 'settings', },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Notification', fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, verbose_name='ID', serialize=False, primary_key=True)
                ), (
                    'message',
                    models.TextField()
                ), (
                    'url',
                    models.CharField(
                        verbose_name='link for notification', null=True, max_length=200, blank=True)
                ), (
                    'is_viewed',
                    models.BooleanField(
                        default=False
                    )
                ), (
                    'is_emailed',
                    models.BooleanField(
                        default=False
                    )
                ), (
                    'created',
                    models.DateTimeField(
                        auto_now_add=True
                    )
                ), (
                    'occurrences',
                    models.PositiveIntegerField(
                        help_text='If the same notification was fired multiple times with no intermediate notifications',
                        default=1,
                        verbose_name='occurrences'
                    )
                ),
            ], options={
                'verbose_name': 'notification', 'db_table': 'nyt_notification', 'ordering': (
                    '-id',), 'verbose_name_plural': 'notifications',
            },
            bases=(models.Model,),
        ),
    ]



from django.db.models import Model
from django.utils.translation import gettext as _

from . import _disable_notifications, models, settings


def notify(message, key, target_object=None, url=None, filter_exclude={}, recipient_users=None):
    """
    Notify subscribing users of a new event. Key can be any kind of string,
    just make sure to reuse it where applicable.

    notify("there was a response to your comment", "comment_response",
           target_object=PostersObject,
           url=reverse('comments:view', args=(PostersObject.id,)))

    The below example notifies everyone subscribing to the "new_comments" key
    with the message "New comment posted".

    notify("New comment posted", "new_comments")

    filter_exclude: a dictionary to exclude special elements of subscriptions
    in the queryset, for instance filter_exclude={''}

    :param: key <string>: Some identifier of your notification type
    :param: recipient_users: A possible iterable of users that should be notified
                             instead of notifying all subscribers of the event.
                             Notice that users still have to be actually subscribed
                             to the event key!
    :param: target_object: Any django model instance that this notification
                           relates to. User django content types.
    """

    if _disable_notifications:
        return 0

    if target_object:
        if not isinstance(target_object, Model):
            raise TypeError(
                _("You supplied a target_object that's not an instance of a django Model."))
        object_id = target_object.id
    else:
        object_id = None

    objects = models.Notification.create_notifications(
        key,
        object_id=object_id,
        message=message,
        url=url,
        filter_exclude=filter_exclude,
        recipient_users=recipient_users,
    )

    # Notify channel subscribers if we have channels enabled
    if settings.ENABLE_CHANNELS:
        from django_nyt import subscribers
        subscribers.notify_subscribers(objects, key)

    return len(objects)


def subscribe(settings, key, content_type=None, object_id=None, **kwargs):
    """
    Creates a new subscription to a given key. If the key does not exist
    as a NotificationType, it will be created

    Uses get_or_create to avoid double creation
    See: https://docs.djangoproject.com/en/dev/ref/models/querysets/#get-or-create

    :param: settings: A models.Settings instance
    :param: key: The unique key that the Settings should subscribe to
    :param: content_type: If notifications are regarding a specific ContentType, it should be set
    :param: object_id: If the notifications should only regard a specific object_id
    :param: **kwargs: Additional models.Subscription field values
    """
    notification_type = models.NotificationType.get_by_key(key, content_type=content_type)

    return models.Subscription.objects.get_or_create(
        settings=settings,
        notification_type=notification_type,
        object_id=object_id,
        **kwargs
    )[0]



from django.conf import settings as django_settings
from django.utils.translation import gettext_lazy as _

DB_TABLE_PREFIX = 'nyt'

ENABLED = getattr(django_settings, 'NYT_ENABLED', True)
"""Global setting to force-fully disable all propagation and creation of
notifications."""

ENABLE_ADMIN = getattr(django_settings, 'NYT_ENABLE_ADMIN', False)
"""Enable django-admin registration for django-nyt's ModelAdmin's"""

SEND_EMAILS = getattr(django_settings, 'NYT_SEND_EMAILS', True)
"""Email notifications global setting, can be used to globally switch off
emails, both instant and scheduled digests.
Remeber that emails are sent with ``python manage.py notifymail``."""

EMAIL_SUBJECT = getattr(
    django_settings,
    'NYT_EMAIL_SUBJECT',
    _("You have new notifications")
)
"""Subject of all emails sent"""

EMAIL_SENDER = getattr(
    django_settings,
    'NYT_EMAIL_SENDER',
    "notifications@example.com"
)
"""Default sender email"""

# You can always make up more numbers... they simply identify which notifications
# to send when invoking the script, and the number indicates how many hours
# to minimum pass between each notification.
INSTANTLY = 0
# Subtract 1, because the job finishes less than 24h before the next...
DAILY = (24 - 1) * 60
WEEKLY = 7 * (24 - 1) * 60

INTERVALS = getattr(
    django_settings,
    'NYT_INTERVALS',
    [
        (INSTANTLY, _('instantly')),
        (DAILY, _('daily')),
        (WEEKLY, _('weekly'))
    ]
)
"""List of intervals available for user selections. In minutes"""

INTERVALS_DEFAULT = INSTANTLY
"""Default selection for new subscriptions"""

USER_MODEL = getattr(django_settings, 'AUTH_USER_MODEL', 'auth.User')


############
# CHANNELS #
############

ENABLE_CHANNELS = (
    'channels' in django_settings.INSTALLED_APPS and
    not getattr(django_settings, 'NYT_CHANNELS_DISABLE', False)
)
"""Channels are enabled automatically when 'channels' application is installed,
however you can explicitly disable it with NYT_CHANNELS_DISABLE."""

# Name of the global channel (preliminary stuff) that alerts everyone that there
# is a new notification
NOTIFICATION_CHANNEL = "nyt_all-{notification_key:s}"


####################
# PLANNED SETTINGS #
####################

# AUTO_DELETE = getattr(django_settings, 'NYT_AUTO_DELETE', 120)
"""After how many days should viewed notifications be deleted? (not implemented)"""

# AUTO_DELETE_ALL = getattr(django_settings, 'NYT_AUTO_DELETE_ALL', 120)
"""After how many days should both viewed and unviewed notifications be deleted? (not implemented)"""



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_nyt', '0005__v_0_9_2'),
    ]

    operations = [
        migrations.AlterField(
            model_name='subscription',
            name='latest',
            field=models.ForeignKey(related_name='latest_for', verbose_name='latest notification', blank=True, to='django_nyt.Notification', null=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
    ]



import logging
import os
import smtplib
import sys
import time
from datetime import datetime

from django.conf import settings
from django.contrib.sites.models import Site
from django.core import mail
from django.core.management.base import BaseCommand
from django.template.loader import render_to_string
from django.utils.translation import activate, deactivate, gettext as _
from django_nyt import models, settings as nyt_settings

# Daemon / mail loop sleep between each database poll (seconds)
SLEEP_TIME = 120


class Command(BaseCommand):
    can_import_settings = True
    # @ReservedAssignment
    help = (
        'Sends notification emails to subscribed users taking into account '
        'the subscription interval'
    )

    def add_arguments(self, parser):
        parser.add_argument(
            '--daemon', '-d',
            action='store_true',
            dest='daemon',
            help='Go to daemon mode and exit'
        )
        parser.add_argument(
            '--cron', '-c',
            action='store_true',
            dest='cron',
            help='Do not loop, just send out emails once and exit'
        )
        parser.add_argument(
            '--pid-file',
            action='store',
            dest='pid',
            help='Where to write PID before exiting',
            default='/tmp/nyt_daemon.pid'
        )
        parser.add_argument(
            '--log-file',
            action='store',
            dest='log',
            help='Where daemon should write its log',
            default='/tmp/nyt_daemon.log'
        )
        parser.add_argument(
            '--no-sys-exit',
            action='store_true',
            dest='no_sys_exit',
            help='Skip sys-exit after forking daemon (for testing purposes)'
        )
        parser.add_argument(
            '--daemon-sleep-interval',
            action='store',
            dest='sleep_time',
            help='Minimum sleep between each polling of the database.',
            default=SLEEP_TIME
        )

    def _send_user_notifications(self, context, connection):
        subject = _(nyt_settings.EMAIL_SUBJECT)

        message = render_to_string(
            'emails/notification_email_message.txt',
            context
        )
        email = mail.EmailMessage(
            subject, message, nyt_settings.EMAIL_SENDER,
            [context['user'].email], connection=connection
        )
        self.logger.info("Sending to: %s" % context['user'].email)
        email.send(fail_silently=False)

    def _daemonize(self):
        self.logger.info("Daemon mode enabled, forking")
        try:
            fpid = os.fork()
            if fpid > 0:
                # Running as daemon now. PID is fpid
                self.logger.info("PID: %s" % str(fpid))
                pid_file = open(self.options['pid'], "w")
                pid_file.write(str(fpid))
                pid_file.close()
                if not self.options['no_sys_exit']:
                    sys.exit(0)
        except OSError as e:
            sys.stderr.write(
                "fork failed: %d (%s)\n" %
                (e.errno, e.strerror))
            sys.exit(1)

    def handle(self, *args, **options):  # noqa: max-complexity=12
        # activate the language
        activate(settings.LANGUAGE_CODE)

        options.setdefault('daemon', False)
        options.setdefault('cron', False)
        options.setdefault('no_sys_exit', False)

        self.options = options

        daemon = options['daemon']
        cron = options['cron']

        assert not (daemon and cron), (
            "You cannot both choose cron and daemon options"
        )

        self.logger = logging.getLogger('django_nyt')

        if not self.logger.handlers:
            if daemon:
                handler = logging.FileHandler(filename=options['log'])
            else:
                handler = logging.StreamHandler(self.stdout)
            self.logger.addHandler(handler)
            self.logger.setLevel(logging.INFO)

        self.logger.info("Starting django_nyt e-mail dispatcher")

        if not nyt_settings.SEND_EMAILS:
            print("E-mails disabled - quitting.")
            sys.exit()

        # Run as daemon, ie. fork the process
        if daemon:
            self._daemonize()

        # create a connection to smtp server for reuse
        connection = mail.get_connection()

        if cron:
            self.send_mails(connection)
            return

        if not daemon:
            print("Entering send-loop, CTRL+C to exit")
        try:
            self.send_loop(connection, int(options['sleep_time']))
        except KeyboardInterrupt:
            print("\nQuitting...")

        # deactivate the language
        deactivate()

    def send_loop(self, connection, sleep_time):

        # This could be /improved by looking up the last notified person
        last_sent = None

        while True:

            started_sending_at = datetime.now()
            self.logger.info(
                "Starting send loop at %s" %
                str(started_sending_at))
            if last_sent:
                user_settings = models.Settings.objects.filter(
                    interval__lte=(
                        (started_sending_at -
                         last_sent).seconds //
                        60) //
                    60).order_by('user')
            else:
                user_settings = None

            self.send_mails(
                connection,
                last_sent=last_sent,
                user_settings=user_settings
            )

            connection.close()
            last_sent = datetime.now()
            elapsed_seconds = (last_sent - started_sending_at).seconds
            time.sleep(
                max(
                    (min(nyt_settings.INTERVALS)[0] - elapsed_seconds) * 60,
                    sleep_time,
                    0
                )
            )

    def _send_batch(self, context, connection, setting):
        """
        Loops through emails in a list of notifications and tries to send
        to each recepient

        """
        # STMP connection send loop
        notifications = context['notifications']

        if len(context['notifications']) == 0:
            return

        while True:
            try:
                self._send_user_notifications(context, connection)
                for n in notifications:
                    n.is_emailed = True
                    n.save()
                break
            except smtplib.SMTPSenderRefused:
                self.logger.error(
                    (
                        "E-mail refused by SMTP server ({}), "
                        "skipping!"
                    ).format(setting.user.email))
                continue
            except smtplib.SMTPException as e:
                self.logger.error(
                    (
                        "You have an error with your SMTP server "
                        "connection, error is: {}"
                    ).format(e))
                self.logger.error("Sleeping for 30s then retrying...")
                time.sleep(30)
            except Exception as e:
                self.logger.error(
                    (
                        "Unhandled exception while sending, giving "
                        "up: {}"
                    ).format(e))
                raise

    def send_mails(self, connection, last_sent=None, user_settings=None):
        """
        Does the lookups and sends out email digests to anyone who has them
        due.
        """

        self.logger.debug("Entering send_mails()")

        connection.open()

        if not user_settings:
            user_settings = models.Settings.objects.all().order_by('user')

        context = {'user': None,
                   'username': None,
                   'notifications': None,
                   'digest': None,
                   'site': Site.objects.get_current()}

        for setting in user_settings:
            context['user'] = setting.user
            context['username'] = getattr(
                setting.user, setting.user.USERNAME_FIELD)
            # Which notifications are remaining for the user's settings
            context['notifications'] = []
            # get the index of the tuple corresponding to the interval and
            # get the string name
            idx = [y[0] for y in nyt_settings.INTERVALS].index(
                setting.interval)
            context['digest'] = nyt_settings.INTERVALS[idx][1]
            for subscription in setting.subscription_set.filter(
                send_emails=True,
                latest__is_emailed=False
            ):
                context['notifications'].append(subscription.latest)

            self._send_batch(context, connection, setting)



from setuptools import find_packages
from setuptools import setup

from django_nyt import __version__


packages = find_packages()

setup(
    name="django-nyt",
    version=__version__,
    author="Benjamin Bach",
    author_email="benjamin@overtag.dk",
    url="https://github.com/benjaoming/django-nyt",
    description="A pluggable notification system written for the Django framework.",
    license="Apache License 2.0",
    keywords=["django", "notification" "alerts"],
    packages=find_packages(),
    zip_safe=False,
    install_requires=["django>=1.11,<2.2"],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
    ],
    include_package_data=True,
)
